# Place a 'nodata fill' border of pixels around an existing Sentinel-2
# image tile. The border thickness can be specified. These border pixels
# are set to a value of 0, the same value as the 'nodata' pixels when
# Sentinel-2 partially images a tile area. This script is intended to be
# run when a Sentinel-2 tile area is completely imaged.

# DISCLAIMER:
# This program DOES NOT change pixel values in an existing image -- it
# simply places a border of "0 pixels" around an image ONLY FOR PURPOSES
# OF IMAGE DISPLAY. Users who might want to perform analyses using bordered
# Sentinel-2 image data generated by this script are assumed/expected to be
# responsible for applying any required radiometric offsets to TOAR and/or
# Sen2Cor-processed image data BEFORE RUNNING THIS SCRIPT--no such adjustment
# is performed in this script.
# CMAC processing accounts for TOAR radiometric offsets, so no adjustment is
# required for this data source.

from pathlib import Path

import cmac_gdal_io as io
import add_border_to_band as border



def main():

    # Initializations:
    
    # Specify the "thickness" of the border in both dimensions.
    total_border_thickness = (5, 5)

    # Consistent with Sentinel .SAFE naming conventions for
    # image data files. CMAC processing assumes/expects this form
    # for its image file basenames.
    image_basename = 'T14TLS_20220712T173921'

    # Final input and output image data paths (not including
    # the name of the image file). This should allow location of
    # Sentinel .SAFE-formatted TOAR/Sen2Cor-processed
    # image data in addition to "loose" image data of any
    # type.
    # These MUST be specified!
    full_input_file_path = 'D:/AddS2NodataBorder_TestData/' \
        + 'S2A_MSIL2A_20220712T173921_N9999_R098_T14TLS_20240127T182829.SAFE/' \
        + 'GRANULE/L2A_T14TLS_A036846_20220712T174933/IMG_DATA/R20m'

    # Strongly recommend having an output path with a folder name representing
    # the type of image considered here (TOA/Sen2Cor/CMAC).
    full_output_file_path = 'D:/AddS2NodataBorder_TestData/Sen2Cor_output'

    # Specify the band(s) to be processed and specify which "type" of image
    # to process. Allowed values for the image type are all upper-case or
    # all lower-case versions of the following:
    # 'TOA', 'TOAR', 'Sen2Cor', 'S2Cor', 'S2C', 'CMAC'
    bands_to_process = ['B05', 'B06', 'B07', 'B8A']
    image_type = 'S2C'

    # Sanity-check the specified image type. The image type string
    # must be one of the all upper-case or all lower-case substrings
    # defined above.
    if not (image_type in ['TOA', 'TOAR', 'toa', 'toar'] \
        or image_type in ['SEN2COR', 'S2COR', 'sen2cor', 's2cor', 'S2C', 's2c'] \
        or image_type in ['CMAC', 'cmac']):
        err_msg = f'ERROR: invalid image type {image_type}'
        print(err_msg)
        return -1
 
    # MAIN PROCESSING LOOP:

    # This script assumes processing of .JP2-formatted input images.
    # These cannot be created directly using GDAL functionality, but
    # they can be created by copying from a GeoTIFF source.
    GTiff_driver = io.gdal.GetDriverByName('GTiff')
    jp2_driver = io.gdal.GetDriverByName('')
    for band in bands_to_process:

        resolution = ''
        if image_type in ['TOA', 'TOAR', 'toa', 'toar']:
        
            image_filename = full_input_file_path + '/' + image_basename \
                + f'_{band}.jp2'

        elif image_type in ['SEN2COR', 'S2COR', 'S2C', 'sen2cor', 's2cor', \
            's2c']:

            # Bordering currently known to work for 10m image data
            # (for use in RGB / NRG color display, obviously).
            # If using .SAFE-formatted Sen2Cor-corrected image data,
            # make sure the "full_input_file_path" parameter includes
            # the "R10m" folder (these bands are initially processed at
            # 20m resolution, and are included in the "R20m" folder).
            if band in ['B02', 'B03', 'B04', 'B08']:
                resolution = '10m'

            # WARNING: As mentioned above, this bordering code was not
            #          initially intended/considered for use
            #          with these bands, so specify at your own risk.
            #          Intial development testing of the script on
            #          TOA(R) .SAFE-formatted image data for these bands
            #          appears to show it works as expected.
            #          If using .SAFE-formatted Sen2Cor-corrected
            #          image data, make sure the "full_input_file_path"
            #          parameter includes the "R20m" folder.
            elif band in ['B05', 'B06', 'B07', 'B8A', 'B11', 'B12']:
                resolution = '20m'

            else:
                # Bordering currently not supported for 60m Coastal-Aerosol
                # (B1), Water Vapor (B9) or Cirrus (B10) band image data.
                # Specifying these bands will display an error message and
                # skip to the next specified band.
                err_msg = f'ERROR: invalid band designator {band}, skipping'
                print(err_msg)
                continue

            image_filename = full_input_file_path + '/' \
                + image_basename + f'_{band}_{resolution}.jp2'

        elif image_type in ['CMAC', 'cmac']:
            
            image_filename = full_input_file_path + '/' + 'CMAC_' \
                + image_basename + f'_{band}.jp2'

         

        # Generate the border 'template' image.
        print(f'Generating {band} template border image...')
        border_filename_GTiff \
            = border.generate_border_image_template(GTiff_driver, image_filename,
                                                    total_border_thickness,
                                                    full_output_file_path,
                                                    image_basename, band,
                                                    'black_border_only')
        if border_filename_GTiff == '':
            print(f'ERROR: cannot create {band} border image template')
            continue

        # Update the "template".
        print(f'Adding {band} image data to template border image...')
        status = border.center_image_data(image_filename, border_filename_GTiff,
                                          total_border_thickness)
        if status < 0:
            print(f'ERROR: Cannot add {band} image to template border image')
            return -1
            

        # Create a JP2 version of the file.
        print(f'Generating "JP2" version of {band} border image...')
        border_filename_jp2 = full_output_file_path + '/' + image_basename \
            + f'_{band}_border_image.jp2'
        status = io.write_image_to_jp2(border_filename_GTiff, border_filename_jp2)
        if status < 0:
            print(f'ERROR: could not generate {band} JP2 version of border image')


        # Remove the GeoTiff version of the file to save space.
        Path(border_filename_GTiff).unlink()


    return 0



if __name__ == '__main__':
    main()

