from pathlib import Path
import subprocess

import numpy as np
from osgeo import osr

import cmac_gdal_io as io



def add_border_to_image(input_image,
                        border_thickness):

    # Get the dimensions of the input image.
    input_dims = input_image.shape

    # Create a "super image" containing 'nodata' fill
    # around its border.
    border_image_dims = ((input_dims[0] + 2*border_thickness[0]), \
        (input_dims[1] + 2*border_thickness[1]))
    bordered_image = np.zeros(border_image_dims, image_band_data.dtype)

    # "Center" the input image data into the bordered output image.
    row_start = border_thickness[0]
    row_end = row_start + image_dims[0]
    col_start = border_thickness[1]
    col_end = col_start + image_dims[1]

    border_image[row_start:row_end, col_start:col_end] = input_image
 
    # Done.
    return bordered_image




def generate_image_extent(input_geotransform,
                          input_image_dims,
                          image_border_thickness):


    # Determine the input image extent (defined as UL and LR corner northing/easting
    # coordinates with units assumed to be standared meters).
    #
    # WARNING: Sentinel-2 image tile MUST be in a UTM-based projection. No guarantees
    #          this will work if the tile is in the lat/lon "projection" represented
    #          by EPSG:4326.
    in_ulx = input_geotransform[0] + 0*input_geotransform[1] + 0*input_geotransform[2]
    in_uly = input_geotransform[3] + 0*input_geotransform[4] + 0*input_geotransform[5]

    in_llx = in_ulx
    #in_llx = input_geotransform[0] + 0*input_geotransform[1] \
    #    + input_image_dims[0]*input_geotransform[2]
    in_lly = input_geotransform[3] + 0*input_geotransform[4] \
        + input_image_dims[0]*input_geotransform[5]

    in_lrx = input_geotransform[0] + input_image_dims[1]*input_geotransform[1] \
        + input_image_dims[0]*input_geotransform[2]
    #in_lry = input_geotransform[3] + input_image_dims[1]*input_geotransform[4] \
    #    + input_image_dims[0]*input_geotransform[5]
    in_lry = in_lly

    #in_urx = input_geotransform[0] + input_image_dims[1]*input_geotransform[1] \
    #    + 0*input_geotransform[2]
    in_urx = in_lrx
    in_ury = in_uly
    #in_ury = input_geotransform[3] + input_image_dims[0]*input_geotransform[4] \
    #    + 0*input_geotransform[5]

   
    # Determine the output image extent. Assume we want to put the input
    # image data into the "center" of the expanded output image. The
    # UL corner of the output image will be shifted up and to the left
    # (i.e. smaller easting / larger northing coordinate), while the
    # LR corner of the output image will be shifted down and to the right
    # (i.e. larger easting / smaller northing coordinate); the shift is
    # based on the border thickness and the pixel spatial resolution.

    # WARNING: This should be strictly true for images processed as
    # "north up"; no guarantee is made for non "north up" images.
    out_ulx = in_ulx - image_border_thickness[1]*input_geotransform[1]
    out_uly = in_uly - image_border_thickness[0]*input_geotransform[5]

    out_lrx = in_lrx + image_border_thickness[1]*input_geotransform[1]
    out_lry = in_lry + image_border_thickness[0]*input_geotransform[5]

    #out_ulx = in_ulx - image_border_thickness[1]*input_geotransform[1]
    #out_uly = in_uly + image_border_thickness[0]*input_geotransform[5] 

    #out_lrx = in_lrx + image_border_thickness[1]*input_geotransform[1]
    #out_lry = in_lry + image_border_thickness[0]*input_geotransform[5]

    return ((out_uly, out_ulx), (out_lry, out_lrx))




def generate_border_image_template(driver, input_image_fname,
                                   border_thickness_dims,
                                   output_border_image_path,
                                   image_basename, band_designator,
                                   border_type='black_border_only'):

    # Get basic info from the input image.
    input_ds = io.open_dataset(input_image_fname, io.gdalconst.GA_ReadOnly)
    if input_ds is None:
        print(f'ERROR: cannot open "{input_image_fname}"')

    input_image_dims = (io.get_num_raster_rows(input_ds), \
        io.get_num_raster_cols(input_ds))
    input_geotransform = io.get_geotransform_info(input_ds)
    input_proj_info = io.get_projection_info(input_ds)

    srs = osr.SpatialReference(wkt=input_proj_info)

    # Generate the extent for the output "border" image.
    print('Determining extent of bordered image...')
    output_extent = generate_image_extent(input_geotransform,
                                          input_image_dims,
                                          border_thickness_dims)

    # Create a "black border" covering the largest spatial extent. 
    black_output_dims_y = input_image_dims[0] + 2*border_thickness_dims[0]
    black_output_dims_x = input_image_dims[1] + 2*border_thickness_dims[1]

    ulx = output_extent[0][1]
    uly = output_extent[0][0]
    lrx = output_extent[1][1]
    lry = output_extent[1][0]

    # Create the output image path if it doesn't exist.
    if not Path(output_border_image_path).exists():
        Path(output_border_image_path).mkdir(parents=True, exist_ok=True)

    output_image_filename = output_border_image_path + '/' + image_basename \
        + f'_{band_designator}_{border_type}.tif'
    dtype = io.gdal.GDT_UInt16
    output_ds = driver.Create(output_image_filename, black_output_dims_x,
                              black_output_dims_y, 1, dtype)

    output_ds.SetProjection(srs.ExportToWkt())

    # Use the UL corner northing coordinate to establish the output image
    # geometric transformation. Using the LR corner northing coordinate
    # will position the bordered image BELOW the unbordered one.
    out_geotransform = [ulx, input_geotransform[1], input_geotransform[2], \
        uly, input_geotransform[4], input_geotransform[5]]
    output_ds.SetGeoTransform(out_geotransform)

    if border_type == 'black_border_only':
        output_ds.GetRasterBand(1).Fill(0)
        output_ds = None

    elif border_type == 'black_and_white_borders':
        black_border = np.zeros((black_output_dims_y, black_output_dims_x), dtype=np.uint16)

        white_output_dims_y = input_image_dims[0] + border_thickness_dims[0]
        white_output_dims_x = input_image_dims[1] + border_thickness_dims[1]
        white_border = 15000*np.ones((white_output_dims_y, white_output_dims_x),
                                     dtype=np.uint16)

        white_border_row_start = border_thickness_dims[0]//2
        white_border_col_start = border_thickness_dims[1]//2

        white_border_row_end = black_output_dims_y - border_thickness_dims[0]//2
        white_border_col_end = black_output_dims_x - border_thickness_dims[1]//2
        
        black_border[white_border_row_start:white_border_row_end, \
            white_border_col_start:white_border_col_end] = white_border

        output_ds.GetRasterBand(1).WriteArray(black_border)
        output_ds = None

    else:
        print(f'ERROR: invalid border specification "{border_type}"')
        output_ds = None
        return ''

    return output_image_filename




def center_image_data(input_image_filename, border_image_filename,
                      image_border_thickness):

    # Get the input image data.
    input_ds = io.open_dataset(input_image_filename, io.gdalconst.GA_ReadOnly)
    if input_ds is None:
        print(f'ERROR: cannot read input image "{input_image_filename}"')
        return -1

    input_image_data = io.read_image_from_open_dataset(input_ds)
    input_ds = None

    # Prep the border image.
    output_ds = io.open_dataset(border_image_filename, io.gdalconst.GA_Update)
    if output_ds is None:
        print(f'ERROR: cannot update border image "{border_image_filename}"')
        return -1
    border_image_data = io.read_image_from_open_dataset(output_ds)

    # "Center" the input image data.
    input_dims = input_image_data.shape

    row_start = image_border_thickness[0]
    row_end = row_start + input_dims[0]

    col_start = image_border_thickness[1]
    col_end = col_start + input_dims[1]
    
    border_image_data[row_start:row_end, col_start:col_end] = input_image_data

    # Write the new array back to the border image dataset.
    output_ds.GetRasterBand(1).WriteArray(border_image_data)

    # Done.
    output_ds = None

    return 0


    
    







    
